// Copyright 2008-2009 Omni Development, Inc.  All rights reserved.
//
// This software may only be used and reproduced according to the
// terms in the file OmniSourceLicense.html, which should be
// distributed with this project and can also be found at
// <http://www.omnigroup.com/developer/sourcecode/sourcelicense/>.

#import "OAApplication.h"

#import "OATargetSelection.h"

RCS_ID("$Header: $")

static BOOL OATargetSelection;

@implementation OAApplication

+ (void)initialize;
{
    if (self == [OAApplication class])
        OATargetSelection = [[NSUserDefaults standardUserDefaults] boolForKey:@"OATargetSelection"];
}

BOOL OADebugTargetSelection = NO;
#define DEBUG_TARGET_SELECTION(format, ...) do { \
    if (OADebugTargetSelection) \
        NSLog((format), ## __VA_ARGS__); \
} while (0)

- (BOOL)sendAction:(SEL)theAction to:(id)theTarget from:(id)sender;
{
    if (OATargetSelection) {
        // The normal NSApplication version, sadly, uses internal target lookup for the nil case. It should really call -targetForAction:to:from:.
        if (!theTarget)
            theTarget = [self targetForAction:theAction to:nil from:sender];
    }
    
    return [super sendAction:theAction to:theTarget from:sender];
}

// Just does our portion of the chain.
- (BOOL)applyToResponderChain:(OAResponderChainApplier)applier;
{
    if (![super applyToResponderChain:applier])
        return NO;
    
    id delegate = (id)self.delegate;
    if (delegate && ![delegate applyToResponderChain:applier])
        return NO;
    
    return YES;
}

// Does the full search documented for -targetForAction:to:from:
- (BOOL)applyToFullResponderChain:(OAResponderChainApplier)applier;
{
    // Follow the normal set of fallbacks as documented for this method on NSApplication.  Terminate if the applier stops (which might be due to finding a target or might be due to one of the candidates claiming the action but refusing to do it).
    NSWindow *keyWindow = [self keyWindow];
    if (keyWindow.firstResponder && ![keyWindow.firstResponder applyToResponderChain:applier])
        return NO;
    
    NSWindow *mainWindow = [self mainWindow];
    if (keyWindow != mainWindow) {
        if (mainWindow.firstResponder && ![mainWindow.firstResponder applyToResponderChain:applier])
            return NO;
    }
    
    // This isn't ideal since this forces an NSDocumentController to be created.  AppKit presumably has some magic to avoid this...  We could avoid this if there are no registered document types, if that becomes an issue.
    NSDocumentController *documentController = [NSDocumentController sharedDocumentController];
    if (documentController && ![documentController applyToResponderChain:applier])
        return NO;
    
    return [self applyToResponderChain:applier];
}

- (id)targetForAction:(SEL)theAction to:(id)theTarget from:(id)sender;
{
    if (!OATargetSelection)
        return [super targetForAction:theAction to:theTarget from:sender];
    
    __block id target = nil;
    
    DEBUG_TARGET_SELECTION(@"looking for target: %@ to:%@ from:%@ (key1:%@, main1:%@)", NSStringFromSelector(theAction), [theTarget shortDescription], [sender shortDescription], [[[self keyWindow] firstResponder] shortDescription], [[[self mainWindow] firstResponder] shortDescription]);
    
    OAResponderChainApplier applier = ^(id object){
        DEBUG_TARGET_SELECTION(@" ... trying %@", [object shortDescription]);
        id responsible = [object responsibleTargetForAction:theAction sender:sender];
        if (responsible) {
            // Someone claimed to be responsible for the action.  The sender will re-validate with any appropriate means and might still get refused, but we should stop iterating.
            target = responsible;
            return NO;
        }
        return YES;
    };
    
    if (theTarget) {
        // The caller had a specific target in mind. In the normal responder chain, this would just be used. Here we are starting at the suggested object and iterating from there. To be more like the default behavior we could swap which line is commented out below.
        [theTarget applyToResponderChain:applier];
        //return theTarget;
    } else
        [self applyToFullResponderChain:applier];
    
    DEBUG_TARGET_SELECTION(@" ... using %@", [target shortDescription]);
    return target;
}

@end
